﻿package classes.scene3d {

	// Import Flash packages
	import flash.display.*;
	import flash.events.*;

	// Import papervision 3D packages
	import org.papervision3d.core.proto.*;
	import org.papervision3d.scenes.*;
	import org.papervision3d.cameras.*;
	import org.papervision3d.objects.*;
	import org.papervision3d.materials.*;
	
	// Import Reflection class
	import com.pixelfumes.*;

	public class ImageGallery extends Sprite {
	
		private var container:Sprite;
		private var scene:MovieScene3D;
		private var camera:FreeCamera3D;
		private var imageCount:Number = 20;
		private var perimeter:Number;
		private var radius:Number;
		private var currentImage:Number = 0;
		private var albums_mc:MovieClip;
		
		/*
		CLASS CONSTRUCTOR
		*/
		public function ImageGallery () {
			
		}
			
		/*
		METHODS
		*/
		public function init(){
			stage.addEventListener(Event.RESIZE, resizeStage);
			init3D();
			addEventListener(Event.ENTER_FRAME, update3D);
		}
		private function init3D() {
			
			// Attach the Albums movie clip to the stage.
			// Note that the library asset needs the Class identifier 'Albums'
			albums_mc = this.addChild(new Albums());
			// Move it off stage so we can't see it
			albums_mc.x = -9999;
			
			// Create container movieclip and center it
			container = new Sprite();
			addChild( container );
			container.x = stage.stageWidth/2;
			container.y = stage.stageHeight/2;
			
			// Create scene
			scene = new MovieScene3D(container);
			
			// Create camera
			camera = new FreeCamera3D();
			camera.x = 0;
			camera.y = 0;
			camera.z = 0;			
			camera.focus = 400;
			camera.zoom = 2;
			
			// Define the size of the circle of images we will be making.
			// The higher the number, the larger the circle.
			perimeter = 6000;
			// From that, work out how large the circle is
			radius = perimeter/Math.PI/2;
			
			// Add each of our album covers to the 3D scene
			for (var n:Number=0;n<imageCount;n++){
				// Add next album cover
				currentImage++;
				addPlane();
			}
			
		}
		private function addPlane(){
			trace("Adding plane "+currentImage);
			// Cleanup our albums mc
			if (currentImage>1){
				// Remove the previous reflection
				albums_mc.removeChildAt(2);
				albums_mc.removeChildAt(1);
			}
			// Send albums mc to the frame with the current album cover
			albums_mc.gotoAndStop(currentImage);
			// Adda reflection to the album cover image
			var myReflect:Reflection = new Reflection(albums_mc, 70, 50);
			// Create a 3D texture from the albums mc
			var material:MovieMaterial = new MovieMaterial(albums_mc,true);
			// Create a new 3D plane, apply the texture to it
			var plane:Plane = new Plane(material, albums_mc.width, albums_mc.height, 1, 5);
			// Position the new plane in the scene: album covers are arranged in a circle.
			var step:Number = (360/imageCount)*(currentImage-1);
			var radian:Number = deg2rad(step);
			plane.x = Math.cos(radian)*radius;
			plane.z = Math.sin(radian)*radius;
			plane.y = -50;
			plane.rotationY = -step+90;
			// Include plane in our 3D scene
			scene.addChild(plane, "Image"+currentImage);
		}
		private function deg2rad(degree):Number {
			// Convert degrees into radians
			return degree*(Math.PI/180);
		}
		private function update3D(event:Event):void {
			// Rotate camera on y axis (simple tweening)
			var camTargYRot:Number = (360/stage.stageWidth*2)*container.mouseX;
			var camDiffYRot:Number = camera.rotationY-camTargYRot;
			camera.rotationY += -camDiffYRot/6;
			// Move camera on y axis (simple tweening)
			var camTargY:Number = (2000/stage.stageHeight)*container.mouseY;
			var camDiffY:Number = camera.y-camTargY;
			camera.y += -camDiffY/6;
			// Rotate camera on x axis (simple tweening)
			var camTargXRot:Number = -(120/stage.stageHeight)*container.mouseY;
			var camDiffXRot:Number = camera.rotationX-camTargXRot;
			camera.rotationX += -camDiffXRot/6;
			// Render the scene
			scene.renderCamera(camera);
		}
		private function resizeStage(event:Event):void {
			container.x = stage.stageWidth/2;
			container.y = stage.stageHeight/2;
		}
	}
	
}